#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <curl/curl.h>
#include <jansson.h>
#include "musicbrainz-lib.h"

struct MBState * mbstate_alloc(void)
{
  struct MBState * mbs ;

  mbs = malloc(sizeof(struct MBState)) ;
  mbs->curl_handle = 0;
  mbs->logged_in = 0;
  return(mbs);
}

void mb_global_initialize()
{
  int cgi_ret;
  /*  curl_global_init(CURL_GLOBAL_WIN32 | CURL_GLOBAL_SSL ) */
  cgi_ret = curl_global_init(CURL_GLOBAL_SSL) ;
  if(cgi_ret != 0)
  {
    printf("curl_global_init() failed\n");
    abort();
  }
}

void mbstate_initialize(struct MBState * mbs)
{
  CURL * ch;

  ch = curl_easy_init() ;
  mbs->curl_handle = ch;
}

struct MBState * mb_initialize(void)
{
  struct MBState * mbs ;

  mbs = mbstate_alloc();
  mbstate_initialize(mbs);

  return(mbs);
}

size_t login_header_callback(char *buffer,
		       size_t size,
		       size_t nitems,
		       void *userdata)
{
  int i;
  return(size * nitems);
}

size_t add_artist_header_callback(char *buffer,
		       size_t size,
		       size_t nitems,
		       void *userdata)
{
  int i;
  return(size * nitems);
}

size_t login_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  return(size * nitems);
}

size_t add_artist_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  return(size * nitems);
}

size_t do_nothing_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  return(size * nitems);
}

size_t build_string_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  char * sofar;
  char * output;
  size_t oldsize, newsize;
  char ** ud;

  ud = (char **) userdata;

  if (userdata == NULL)
    return(size * nitems);

  /*
  sofar = (char *) *userdata;
  sofar = *ud;
  */
  sofar = * (char **) userdata;

  oldsize = (sofar == NULL ? 0 : strlen(sofar));

  newsize = oldsize + size * nitems + 1;
  sofar = realloc(sofar, newsize);
  sofar[oldsize] = 0;
  strncat(sofar, buffer, size * nitems);
  sofar[oldsize + size * nitems] = 0;

  /*
  * (char **) userdata = sofar;
  */
  *ud = sofar;
  return(size * nitems);
}

int nullify_curl_options(CURL * curl, int count, CURLoption options[])
{
  int i;
  for(i = 0; i < count ; i++)
  {
    switch(options[i]) {
    case CURLOPT_WRITEDATA :
      curl_easy_setopt(curl, options[i], stdout);
      break;
    default :
      curl_easy_setopt(curl, options[i], NULL);
    }
  }
}

int mb_login(struct MBState * mbs, char* username, char* password)
{
  CURLcode ret;

  char * postfields;
  int postfields_length;
  int numchars;
  char * header = NULL;
  long int status_code = 0;
  char * redirect_url;

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL,
			 "https://musicbrainz.org/login");
  postfields_length = strlen(username) + strlen(password) + 20 ;
  postfields = malloc(postfields_length) ;

  assert(sprintf(postfields, "username=%s&password=%s", username, password) <
	 postfields_length) ;
 
  /* permanent options */
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_COOKIEFILE, "");

  /* options for just this transfer */
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_POSTFIELDS, postfields) ;
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HEADERDATA, &header) ;
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HEADERFUNCTION,
			 &login_header_callback);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &login_write_callback);

  ret = curl_easy_perform(mbs->curl_handle);

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  /*
  printf("  status_code is %ld\n", status_code);
  printf("  redirect_url is %s\n", redirect_url);
  */

  CURLoption opts[] = {CURLOPT_HEADERFUNCTION, CURLOPT_POSTFIELDS,
			   CURLOPT_HEADERDATA, CURLOPT_WRITEFUNCTION};
  nullify_curl_options(mbs->curl_handle, 4, opts);

  free((void *) postfields);
  if(status_code == 302)
  {
    mbs->logged_in = 1;
    return(0);
  }
  else
  {
    mbs->logged_in = 0;
    return(1);
  }
}

char * build_form_chunk(CURL * ch, char * field_name, char * field_val)
{
  char * field_val_encoded;
  char * out;
  int postfields_length;
  CURLcode ret;

  field_val_encoded = curl_easy_escape(ch, field_val, 0);

  char * postformat = "%s=%s";
  postfields_length = strlen(postformat) + strlen(field_name) +
    strlen(field_val_encoded);
  out = malloc(postfields_length);
  assert(sprintf(out, postformat, field_name, field_val_encoded) <
	 postfields_length) ;

  curl_free(field_val_encoded);

  return(out);
}

char * itoa (int num) {
  int len = 0;
  int x;
  char * out;

  for (x = num; x != 0; x = x / 10)
    len = len + 1;

  if(num < 0)
    len = len + 1;

  out = malloc(len + 1);
  sprintf(out, "%d", num);
  return(out);
  
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
char * add_chunk_to_postfields(char * postfields, char * chunk)
{
  char * postfields2;
  int postfields_length = strlen(postfields);

  postfields_length += (strlen(chunk) + 2) ;

  postfields2 = malloc(postfields_length) ;
  sprintf(postfields2, "%s&%s", postfields, chunk);

  free(postfields);

  return(postfields2);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(*postfields) or a realloc
*/
void add_chunk_to_postfields3(char ** postfields, char * chunk)
{
  char * postfields2;
  int postfields_length = strlen(*postfields);

  postfields_length += (strlen(chunk) + 2) ;

  postfields2 = malloc(postfields_length) ;
  sprintf(postfields2, "%s&%s", *postfields, chunk);

  free(*postfields);

  *postfields = postfields2;
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
char * add_chunk_to_postfields2(CURL * curl, char * postfields,
				char * field_name, char * field_val)
{
  char * newchunk;

  newchunk = build_form_chunk(curl, field_name, field_val);
  postfields = add_chunk_to_postfields(postfields, newchunk);
  /*
  add_chunk_to_postfields3(&postfields, newchunk);
  */
  free(newchunk);

  return(postfields);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(*postfields) or a realloc
*/
void add_chunk_to_postfields4(CURL * curl, char ** postfields,
				char * field_name, char * field_val)
{
  char * newchunk;

  newchunk = build_form_chunk(curl, field_name, field_val);
  *postfields = add_chunk_to_postfields(*postfields, newchunk);
  /*
  add_chunk_to_postfields3(postfields, newchunk);
  */
  free(newchunk);

}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(*postfields) or a realloc
*/
void add_chunk_to_postfields5(CURL * curl, char ** postfields,
				char * field_name, int field_val)
{
  char * newchunk;
  char * field_val_string;

  field_val_string = itoa(field_val);

  newchunk = build_form_chunk(curl, field_name, field_val_string);
  *postfields = add_chunk_to_postfields(*postfields, newchunk);
  /*
  add_chunk_to_postfields3(postfields, newchunk);
  */

  free(field_val_string);
  free(newchunk);

}

int add_numbered_chunk_to_postfields(CURL * curl, char ** postfields,
				     char * field, int num, char * value)
{
  char * field2;

  field2 = malloc(strlen(field) + num);

  sprintf(field2, field, num);

  add_chunk_to_postfields4(curl, postfields, field2, value);

  free(field2);
  
  return(0);
}

int add_numbered_chunk_to_postfields2(CURL * curl, char ** postfields,
				      char * buf,
				      char * field, int num, char * value)
{
  char * field2 = buf;

  sprintf(buf, field, num);

  add_chunk_to_postfields4(curl, postfields, buf, value);
  
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_type_to_artist(CURL *curl, char ** postfields,
		       struct MBArtistInfo artist)
{
  if (artist.type != 0)
    add_chunk_to_postfields5(curl, postfields,
			     "edit-artist.type_id", artist.type);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_gender_to_artist(CURL *curl, char ** postfields,
			 struct MBArtistInfo artist)
{
  if (artist.gender != 0)
    add_chunk_to_postfields5(curl, postfields,
			     "edit-artist.gender_id", artist.gender);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_disambig_to_artist(CURL *curl, char ** postfields,
			  struct MBArtistInfo artist)
{
  if (artist.disambiguation == NULL)
    return(0);

  add_chunk_to_postfields4(curl, postfields,
			   "edit-artist.comment", artist.disambiguation);

  if(artist.not_dupe == 1)
    add_chunk_to_postfields3(postfields, "edit-artist.not_dupe=1");

  return(0);
}

/*
reads area_id field of the "artist" parameter, which is an int, and adds the
appropriate field to postfields.

assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_area_to_artist(CURL *curl, char ** postfields,
			  struct MBArtistInfo artist)
{
  if (artist.area_id != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-artist.area_id", artist.area_id);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_urls_to_artist(CURL *curl, char ** postfields,
			  struct MBArtistInfo artist)
{
  char * lt_field_name;
  char * text_field_name;

  int i;

  char * ltfnf = "edit-artist.url.%d.link_type_id";
  char * tfnf = "edit-artist.url.%d.text";

  lt_field_name = malloc(strlen(ltfnf) + artist.url_count);
  text_field_name = malloc(strlen(tfnf) + artist.url_count);

  for (i = 0; i < artist.url_count && artist.urls; i++)
  {
    sprintf(lt_field_name, ltfnf, i);
    sprintf(text_field_name, tfnf, i);

    add_chunk_to_postfields5(curl, postfields,
			     lt_field_name, artist.urls[i].type);

    add_chunk_to_postfields4(curl, postfields,
			     text_field_name, artist.urls[i].address);
  }

  free(lt_field_name);
  free(text_field_name);
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_relationships_to_artist(CURL *curl, char ** postfields,
				   struct MBArtistInfo artist)
{
  char * bwfield;
  char * typefield;
  char * targetfield;

  int i;

  char * bwfieldformat = "edit-artist.rel.%d.backward";
  bwfield = malloc(strlen(bwfieldformat) + artist.relationship_count);
  char * typefieldformat = "edit-artist.rel.%d.link_type_id";
  typefield = malloc(strlen(typefieldformat) + artist.relationship_count);
  char * targetfieldformat = "edit-artist.rel.%d.target";
  targetfield = malloc(strlen(targetfieldformat) + artist.relationship_count);

  for (i = 0; i < artist.relationship_count && artist.relationships; i++)
  {
    if (artist.relationships[i].backward)
    {
      sprintf(bwfield, bwfieldformat, i);

      add_chunk_to_postfields4(curl, postfields, bwfield, "1");
    }

    sprintf(typefield, typefieldformat, i);

    add_chunk_to_postfields5(curl, postfields,
			     typefield, artist.relationships[i].type);

    sprintf(targetfield, targetfieldformat, i);

    add_chunk_to_postfields4(curl, postfields, targetfield,
			     artist.relationships[i].musicbrainzID);
  }
  free(bwfield);
  free(typefield);
  free(targetfield);

  return(0);
}

/*
This will probably puke/crash if artist_name or sort_name field is NULL
 */
char * build_artist_name_post_string(CURL * ch, struct MBArtistInfo artist)
{
  char * postfields;
  char * sortname_chunk;

  postfields = build_form_chunk(ch, "edit-artist.name", artist.artist_name);

  sortname_chunk = build_form_chunk(ch, "edit-artist.sort_name",
				    artist.sort_name);
  postfields = add_chunk_to_postfields(postfields, sortname_chunk);

  free(sortname_chunk);

  return(postfields);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_disambiguation_to_place(CURL *curl, char ** postfields,
				   struct MBPlace place)
{
  if (place.disambiguation != NULL)
    add_chunk_to_postfields4(curl, postfields,
			     "edit-place.comment", place.disambiguation);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_dupe_to_place(CURL *curl, char ** postfields, struct MBPlace place)
{
  if (place.not_dupe == 1)
    add_chunk_to_postfields4(curl, postfields, "edit-place.not_dupe", "1");

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_type_to_place(CURL *curl, char ** postfields, struct MBPlace place)
{
  if (place.type != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-place.type_id", place.type);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_coordinates_to_place(CURL *curl, char ** postfields,
				struct MBPlace place)
{
  char * coords;

  if (place.coordinates == NULL)
    coords = "";
  else
    coords = place.coordinates;

  add_chunk_to_postfields4(curl, postfields, "edit-place.coordinates", coords);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_address_to_place(CURL *curl, char ** postfields,
			    struct MBPlace place)
{
  if (place.address != NULL)
    add_chunk_to_postfields4(curl, postfields,
			     "edit-place.address", place.address);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_area_id_to_place(CURL *curl, char ** postfields,
			    struct MBPlace place)
{
  if (place.area_id != 0)
    add_chunk_to_postfields5(curl, postfields,
			     "edit-place.area_id", place.area_id);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_relationships_to_place(CURL * curl, char ** postfields,
				  struct MBPlace place)
{
  char * bwfield;
  char * typefield;
  char * targetfield;

  int i;

  char * bwfieldformat = "edit-place.rel.%d.backward";
  bwfield = malloc(strlen(bwfieldformat) + place.relationship_count);
  char * typefieldformat = "edit-place.rel.%d.link_type_id";
  typefield = malloc(strlen(typefieldformat) + place.relationship_count);
  char * targetfieldformat = "edit-place.rel.%d.target";
  targetfield = malloc(strlen(targetfieldformat) + place.relationship_count);

  for (i = 0; i < place.relationship_count && place.relationships; i++)
  {
    if (place.relationships[i].backward)
    {
      sprintf(bwfield, bwfieldformat, i);

      add_chunk_to_postfields4(curl, postfields, bwfield, "1");
    }

    sprintf(typefield, typefieldformat, i);
    add_chunk_to_postfields5(curl, postfields,
			     typefield, place.relationships[i].type);

    sprintf(targetfield, targetfieldformat, i);

    add_chunk_to_postfields4(curl, postfields, targetfield,
			     place.relationships[i].musicbrainzID);
  }
  free(bwfield);
  free(typefield);
  free(targetfield);

  return(0);
}

/*
assumes that place->relationship_count and place->relationships are set to
sane values
*/
int add_relationship_to_place(struct MBPlace * place, MBID relative,
			      int type, int backward)
{
  int rc_new;
  struct MBRelationship * new_relationships;

  rc_new = place->relationship_count + 1;
  new_relationships = realloc(place->relationships,
			      sizeof(struct MBRelationship) * rc_new);
  if (!new_relationships)
  {
    printf("error with realloc in add_relationship_to_place\n");
    abort();
  }

  place->relationships = new_relationships;
  place->relationships[place->relationship_count].type = type;
  place->relationships[place->relationship_count].backward = backward;
  strncpy(place->relationships[place->relationship_count].musicbrainzID,
	  relative, MBID_length);
  place->relationships[place->relationship_count].musicbrainzID[MBID_length] = '\0';
  place->relationship_count = rc_new;

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_urls_to_place(CURL *curl, char ** postfields,
			  struct MBPlace place)
{
  char * ulti_field;
  char * ut_field;

  int i;

  char * ulti_format = "edit-place.url.%d.link_type_id";
  char * ut_format = "edit-place.url.%d.text";

  ulti_field = malloc(strlen(ulti_format) + place.url_count);
  ut_field = malloc(strlen(ut_format) + place.url_count);

  for (i = 0; i < place.url_count && place.urls; i++)
  {
    sprintf(ulti_field, ulti_format, i);
    sprintf(ut_field, ut_format, i);

    add_chunk_to_postfields5(curl, postfields, ulti_field, place.urls[i].type);

    add_chunk_to_postfields4(curl, postfields,
			     ut_field, place.urls[i].address);
  }
  free(ulti_field);
  free(ut_field);

  return(0);
}

char * build_place_name_post_string(CURL * ch, struct MBPlace place)
{
  return(build_form_chunk(ch, "edit-place.name", place.name));
}

char * build_event_name_post_string(CURL * ch, struct MBEventInfo event)
{
  return(build_form_chunk(ch, "edit-event.name", event.name));
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_setlist_to_event(CURL *curl, char ** postfields,
				struct MBEventInfo event)
{
  char * setlist = (event.setlist == NULL ? "" : event.setlist);

  add_chunk_to_postfields4(curl, postfields, "edit-event.setlist", setlist);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_disambig_to_event(CURL *curl, char ** postfields,
			     struct MBEventInfo event)
{
  char * disambig;

  if (event.disambiguation == NULL)
    disambig = "";
  else
    disambig = event.disambiguation;

  add_chunk_to_postfields4(curl, postfields, "edit-event.comment", disambig);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_type_to_event(CURL * curl, char ** postfields,
			 struct MBEventInfo event)
{
  if (event.type != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.type_id", event.type);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_begin_year_to_event(CURL * curl, char ** postfields,
			       struct MBEventInfo event)
{
  if (event.begin_year != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.begin_date.year",
			     event.begin_year);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_begin_month_to_event(CURL * curl, char ** postfields,
			     struct MBEventInfo event)
{
  if (event.begin_month != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.begin_date.month",
			     event.begin_month);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_begin_day_to_event(CURL * curl, char ** postfields,
			      struct MBEventInfo event)
{
  if (event.begin_day != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.begin_date.day",
			     event.begin_day);
  }
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_end_year_to_event(CURL * curl, char ** postfields,
			       struct MBEventInfo event)
{
  if (event.end_year != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.end_date.year",
			     event.end_year);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_end_month_to_event(CURL * curl, char ** postfields,
			     struct MBEventInfo event)
{
  if (event.end_month != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.end_date.month",
			     event.end_month);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_end_day_to_event(CURL * curl, char ** postfields,
			      struct MBEventInfo event)
{
  if (event.end_day != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.end_date.day",
			     event.end_day);
  }
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_time_to_event(CURL * curl, char ** postfields,
			 struct MBEventInfo event)
{
  if (event.time != NULL)
    add_chunk_to_postfields4(curl, postfields, "edit-event.time", event.time);

  return(0);
}

int add_urls_to_event(CURL * curl, char ** postfields,
		      struct MBEventInfo event)
{
  char * lt_field_name;
  char * text_field_name;

  int i;

  char * ltfnf = "edit-event.url.%d.link_type_id";
  char * tfnf = "edit-event.url.%d.text";

  lt_field_name = malloc(strlen(ltfnf) + event.url_count);
  text_field_name = malloc(strlen(tfnf) + event.url_count);

  for (i = 0; i < event.url_count && event.urls; i++)
  {
    sprintf(lt_field_name, ltfnf, i);
    sprintf(text_field_name, tfnf, i);

    add_chunk_to_postfields5(curl, postfields,
			     lt_field_name, event.urls[i].type);

    add_chunk_to_postfields4(curl, postfields,
			     text_field_name, event.urls[i].address);
  }

  free(lt_field_name);
  free(text_field_name);
  return(0);
}

/*
adds relationships from the "relationships" field, as well as the venue, if
given

assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_relationships_to_event(CURL * curl, char ** postfields,
				  struct MBEventInfo event)
{
  int i = 0;
  char * bwfield;
  char * typefield;
  char * targetfield;

  char * bwfieldformat = "edit-event.rel.%d.backward";
  char * typefieldformat = "edit-event.rel.%d.link_type_id";
  char * targetfieldformat = "edit-event.rel.%d.target";
  bwfield = malloc(strlen(bwfieldformat) + event.relationship_count);
  typefield = malloc(strlen(typefieldformat) + event.relationship_count);
  targetfield = malloc(strlen(targetfieldformat) + event.relationship_count);

  for ( ; i < event.relationship_count && event.relationships; i++)
  {
    if (event.relationships[i].backward)
    {
      add_numbered_chunk_to_postfields2(curl, postfields, bwfield,
					bwfieldformat, i, "1");
    }

    sprintf(typefield, typefieldformat, i);
    add_chunk_to_postfields5(curl, postfields,
			     typefield, event.relationships[i].type);

    sprintf(targetfield, targetfieldformat, i);
    add_chunk_to_postfields4(curl, postfields, targetfield,
			     event.relationships[i].musicbrainzID);
  }

  if (strlen(event.venue) == MBID_length)
  {
    sprintf(typefield, typefieldformat, i);
    add_chunk_to_postfields4(curl, postfields, typefield, "794");
    sprintf(targetfield, targetfieldformat, i);
    add_chunk_to_postfields4(curl, postfields, targetfield, event.venue);
  }
  
  return(0);
}

/*
External function (i.e. consider it part of the interface to this library)
Add event referred to by "relative", as a relative, to "ei"
assumes that ei->relationship_count and ei->relationships are set to
sane values
*/
int add_relationship_to_event(struct MBEventInfo * ei, MBID relative,
			      int type, int backward)
{
  int rc_new;
  struct MBRelationship * new_relationships;

  rc_new = ei->relationship_count + 1;
  new_relationships = realloc(ei->relationships,
			      sizeof(struct MBRelationship) * rc_new);
  if (!new_relationships)
  {
    printf("error with realloc in add_relationship_to_event\n");
    abort();
  }

  ei->relationships = new_relationships;
  ei->relationships[ei->relationship_count].type = type;
  ei->relationships[ei->relationship_count].backward = backward;
  strncpy(ei->relationships[ei->relationship_count].musicbrainzID,
	  relative, MBID_length);
  ei->relationships[ei->relationship_count].musicbrainzID[MBID_length] = '\0';
  ei->relationship_count = rc_new;

  return(0);
}

/*
assumes that place->relationship_count and place->relationships are set to
sane values
*/
int add_parent_to_place(struct MBPlace * place, MBID parent_place)
{
  return(add_relationship_to_place(place, parent_place, 717, 1));
}

/*
assumes that ei->relationship_count and ei->relationships are set to
sane values
*/
int add_parent_to_event(struct MBEventInfo * ei, MBID parent_event)
{
  return(add_relationship_to_event(ei, parent_event, 818, 1));
}

int add_series_to_event(struct MBEventInfo * ei, MBID series)
{
  return(add_relationship_to_event(ei, series, 802, 0));
}

char * download_entity_info(struct MBState * mbs, MBID entity)
{
  char * area_url;
  char * entity_url;
  char * resultpage = NULL;
  long int status_code = 0;
  CURLcode ret;
  struct curl_slist * headers = NULL;

  char * entity_url_format = "https://musicbrainz.org/ws/js/entity/%s";

  area_url = malloc(strlen(entity_url_format) + MBID_length);

  sprintf(area_url, entity_url_format, entity);

  entity_url = area_url;

  headers = curl_slist_append(headers, "Accept: application/json");
  if (headers == NULL)
  {
    printf("L1057: something went wrong with curl_slist_append\n");
  }

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HTTPGET, 1);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL, entity_url);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HTTPHEADER, headers);

  ret = curl_easy_perform(mbs->curl_handle);

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);

  curl_slist_free_all(headers);
  curl_easy_reset(mbs->curl_handle);

  return(resultpage);
}

int lookup_area(struct MBState * mbs, MBID area)
{
  char * area_url;
  char * resultpage = NULL;
  long int status_code = 0;
  CURLcode ret;
  struct curl_slist * headers = NULL;
  json_t * data;
  json_t * id_thing;
  json_t * et_thing;
  json_error_t error;
  int id_int;

  resultpage = download_entity_info(mbs, area);

  data = json_loads(resultpage, 0, &error);
  id_thing = json_object_get(data, "id");
  et_thing = json_object_get(data, "entityType");

  if(!json_is_string(et_thing))
  {
    printf("L1099: entityType entry is not a string\n");
    json_decref(data);
    free(resultpage);
    return(-1);
  }

  if(strcmp("area", json_string_value(et_thing)) != 0)
  {
    printf("L1107: not an area\n");
    json_decref(data);
    free(resultpage);
    return(-1);
  }

  switch(json_typeof(id_thing)) {
    case JSON_STRING:
      id_int = atoi(json_string_value(id_thing));
      json_decref(data);
      break;
    case JSON_INTEGER:
      id_int = json_integer_value(id_thing);
      json_decref(data);
      break;
    case JSON_OBJECT:
      json_decref(data);
      free(resultpage);
      return(-1);
    default :
      json_decref(data);
      free(resultpage);
      return(-1);
  }

  free(resultpage);
  return(id_int);
}

char * read_type_from_entity_info(char * entity_info_text)
{
  json_t * data;
  json_t * et_thing;
  json_error_t error;
  int id_int;
  char * type_value;

  if(entity_info_text == NULL)
    return(NULL);

  data = json_loads(entity_info_text, 0, &error);
  et_thing = json_object_get(data, "entityType");

  if(!json_is_string(et_thing))
  {
    printf("L1152: entityType entry is not a string\n");
    json_decref(data);
    return(NULL);
  }

  type_value = malloc(strlen(json_string_value(et_thing)) + 1);
  strcpy(type_value, json_string_value(et_thing));

  json_decref(data);
  return(type_value);
}

/*
returns a pointer to a string in malloc'ed space -- should be free'd when done
 */
char * download_edit_page(CURL * ch, char entity_type[], char entity[])
{
  char * url;
  CURLcode ret;
  char * urlformat = "https://musicbrainz.org/%s/%s/edit";
  char * resultpage = NULL;
  long int status_code = 0;

  url = malloc(strlen(urlformat) + strlen(entity_type) + strlen(entity));
  sprintf(url, urlformat, entity_type, entity);

  ret = curl_easy_setopt(ch, CURLOPT_HTTPGET, 1L);
  ret = curl_easy_setopt(ch, CURLOPT_URL, url);
  ret = curl_easy_setopt(ch, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);
  ret = curl_easy_perform(ch);
  if(ret != CURLE_OK)
  {
    printf("curl_easy_perform returned %s\n", curl_easy_strerror(ret));
  }
  ret = curl_easy_getinfo(ch, CURLINFO_RESPONSE_CODE,
			  &status_code);
  free(url);
  return(resultpage);
}

/* doesn't do any memory-management stuff relating to "data" input value */
/* puts a pointer to a malloc'ed array of relationships in memory pointed by
   relout -- free it when done with it */
int read_relationships_from_json(json_t * data,
				 struct MBRelationship ** relout,
				 struct MBRelationshipInfo ** relinfoout)
{
  struct MBRelationship * relationships;
  struct MBRelationshipInfo * relinfo;

  int i;
  int found_count = 0;

  json_t * rel_json_item;
  json_t * rel_target_json_item;
  json_t * rel_id;
  json_t * rel_type;
  json_t * rel_target;
  json_t * rel_dir;
  json_t * rel_name;
  json_t * target_name;

  const char * json_string;

  if (!json_is_array(data))
  {
    *relout = NULL;
    return(0);
  }
  relationships = malloc(sizeof(struct MBRelationship)* json_array_size(data));
  relinfo = malloc(sizeof(struct MBRelationshipInfo)* json_array_size(data));

  for(i = 0; i < json_array_size(data) ; i++)
  {
    rel_json_item = json_array_get(data, i);
    if(!json_is_object(rel_json_item))
      continue;
    rel_target_json_item = json_object_get(rel_json_item, "target");
    if(!json_is_object(rel_target_json_item))
      continue;
    rel_id = json_object_get(rel_json_item, "id");
    rel_type = json_object_get(rel_json_item, "linkTypeID");
    rel_target = json_object_get(rel_target_json_item, "gid");
    if(json_is_integer(rel_id) && json_is_integer(rel_type) &&
       json_is_string(rel_target))
    {
      relationships[found_count].id = json_integer_value(rel_id);
      relationships[found_count].type = json_integer_value(rel_type);
      strncpy(relationships[found_count].musicbrainzID,
	      json_string_value(rel_target),
	      MBID_length);
      relationships[found_count].musicbrainzID[MBID_length] = '\0';
      rel_dir = json_object_get(rel_json_item, "direction");
      relationships[found_count].backward = 0;
      if ( json_is_string(rel_dir) &&
	   strcmp(json_string_value(rel_dir), "backward") == 0)
	relationships[found_count].backward = 1;
      rel_name = json_object_get(rel_json_item, "verbosePhrase");
      target_name = json_object_get(rel_target_json_item, "name");
      relinfo[found_count].id = relationships[found_count].id;
      if (json_is_string(rel_name))
      {
	json_string = json_string_value(rel_name);
	relinfo[found_count].rel_name = malloc(strlen(json_string) + 1);
	strcpy(relinfo[found_count].rel_name, json_string);
      }
      if (json_is_string(target_name))
      {
	json_string = json_string_value(target_name);
	relinfo[found_count].target_name = malloc(strlen(json_string) + 1);
	strcpy(relinfo[found_count].target_name, json_string);
      }
      found_count = found_count + 1;
    }
  }
  relationships = realloc(relationships,
			  sizeof(struct MBRelationship) * found_count);
  relinfo = realloc(relinfo, sizeof(struct MBRelationshipInfo) * found_count);
  *relout = relationships;
  *relinfoout = relinfo;
  return(found_count);
}

/* puts a pointer to a malloc'ed array of relationships in memory pointed by
   relout -- free it when done with it */
int pull_relationships_from_edit_page(char * pagetext,
				      struct MBRelationship ** relationships,
				      struct MBRelationshipInfo ** relinfo
				      )
{
  char * relstart;
  char * jsontext1;
  char * jsontext;
  int relcount;

  json_t * data;
  int i = 0;
  int inquote = 0;
  relstart = strstr(pagetext, "MB.initRelationshipEditors");

  if (relstart == NULL)
  {
    *relationships = NULL;
    return(0);
  }
  while(1) {
    if (relstart[i] == '\0')
      break;
    if (inquote == 1)
    {
      if (relstart[i] == '\\')
      {
	i = i + 2;
      }
      else
      {
	if (relstart[i] == '"')
	{
	  inquote = 0;
	}
	i = i + 1;
      }
    }
    else
    {
      if (relstart[i] == ';')
	break;
      if (relstart[i] == '"')
      {
	inquote = 1;
      }
      i = i + 1;
    }
  };

  jsontext1 = malloc(i+1);
  strncpy(jsontext1, relstart, i);
  jsontext1[i] = '\0';

  json_error_t error;

  jsontext = strchr(strstr(jsontext1, "relationships"), '[');
  data = json_loads(jsontext, JSON_DISABLE_EOF_CHECK, &error);
  relcount = read_relationships_from_json(data, relationships, relinfo);

  json_decref(data);
  free(jsontext1);
  return(relcount);

}

/* return value is the number of relationships found */
int get_relationship_info(CURL * ch, char entity_type[], char entity[],
			  struct MBRelationship ** relout,
			  struct MBRelationshipInfo ** relinfoout
			  )
{
  char * pagetext;
  FILE * pageout;
  int relcount;
  /*
  struct MBRelationship * relationships;
  struct MBRelationshipInfo * relinfo;
  */

  pagetext = download_edit_page(ch, entity_type, entity);
  relcount = pull_relationships_from_edit_page(pagetext, relout, relinfoout);
  /*
  relationships = *relout;
  relinfo = *relinfoout;
  */

  free(pagetext);
  return(relcount);
}

void mb_init_ArtistInfo(struct MBArtistInfo * ai)
{
  /*
  ai->artist_name = ai->sort_name = ai->disambiguation = ai->urls = NULL;
  */
  ai->artist_name = NULL;
  ai->sort_name = NULL;
  ai->disambiguation = NULL;
  ai->urls = NULL;
  ai->area_id = 0;
  ai->type = Unknown;
  ai->gender = genderUnknown;
  ai->relationships = NULL;
  ai->not_dupe = ai->url_count = ai->relationship_count = 0;
}

void mb_init_Place(struct MBPlace * p)
{
  /*
  p->artist_name = p->sort_name = p->disambiguation = p->urls = NULL;
  */
  p->name = NULL;
  p->disambiguation = NULL;
  p->type = placeUnknown;
  p->coordinates = NULL;
  p->address = NULL;
  p->area_id = 0;
  p->urls = NULL;
  p->relationships = NULL;
  p->not_dupe = p->url_count = p->relationship_count = 0;
}

void mb_init_EventInfo(struct MBEventInfo * ei)
{
  ei->name = NULL;
  ei->disambiguation = NULL;
  ei->setlist = NULL;
  ei->type = 0;
  ei->begin_year = 0;
  ei->begin_month = 0;
  ei->begin_day = 0;
  ei->end_year = 0;
  ei->end_month = 0;
  ei->end_day = 0;
  ei->time = NULL;
  ei->venue[0] = '\0';
  ei->url_count = 0;
  ei->urls = NULL;
  ei->relationship_count = 0;
  ei->relationships = NULL;
}

int get_mbid_from_artist_url(char * url, MBID * mbidp)
{
  int slash;
  char * url2;
  char * url3;
  int i;

  if(url == NULL)
  {
    (*mbidp)[0] = '\0';
    return(0);
  }

  url2 = malloc(strlen(url) + 1);
  strcpy(url2, url);
  
  for(i = strlen(url2) - 1 ; url2[i] == '/'; i--)
    url2[i] = 0;

  url3 = strrchr(url2, '/') + 1;

  strncpy(*mbidp, url3, MBID_length);
  (*mbidp)[MBID_length] = 0;
  free(url2);
}

/* mbidp must be a pointer to allocated space, that mb_send_post_query will
   write into, or NULL if you don't want it written into.
*/
int mb_send_post_query(CURL * ch, char * url, char * postfields,
		       MBID * mbidp)
{
  int ret;
  char * resultpage = NULL;
  long int status_code = 0;
  char * redirect_url;

  ret = curl_easy_setopt(ch, CURLOPT_URL, url);

  ret = curl_easy_setopt(ch, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);

  ret = curl_easy_setopt(ch, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(ch);

  CURLoption opts[] = {CURLOPT_WRITEDATA, CURLOPT_WRITEFUNCTION,
		       CURLOPT_POSTFIELDS};
  switch(ret) {
    case CURLE_OK:
      break;
    case CURLE_COULDNT_RESOLVE_HOST:
      free(resultpage);
      /*      printf("L1476: couldn't resolve host\n"); */
      nullify_curl_options(ch, 3, opts);
      return(1);
    default:
      printf("L1480: curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(ch, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(ch, CURLINFO_REDIRECT_URL,
			  &redirect_url);
  /*
  printf("  L1488: status_code is %ld\n", status_code);
  printf("  L1489: redirect_url is %s\n", redirect_url);
  */

  if(status_code == 200 || status_code == 500)
  {
    nullify_curl_options(ch, 3, opts);
    free(resultpage);
    return(1);
  }

  if(mbidp != NULL)
  {
    get_mbid_from_artist_url(redirect_url, mbidp);
  }

  nullify_curl_options(ch, 3, opts);

  free(resultpage);
  return(0);
}

/*
name and sort_name fields of "artist" argument are required
*/
int mb_do_artist(CURL * ch, char * url, struct MBArtistInfo artist)
{
  char * postfields;
  int ret;

  MBID artist_mbid;
  MBID * artist_mbidp;
  artist_mbidp = &artist_mbid;

  postfields = build_artist_name_post_string(ch, artist);

  add_disambig_to_artist(ch, &postfields, artist);
  add_type_to_artist(ch, &postfields, artist);
  add_gender_to_artist(ch, &postfields, artist);
  add_area_to_artist(ch, &postfields, artist);
  add_urls_to_artist(ch, &postfields, artist);
  add_relationships_to_artist(ch, &postfields, artist);

  ret = mb_send_post_query(ch, url, postfields, artist_mbidp);

  free((void *) postfields);
  return(0);
}

int mb_add_artist(struct MBState * mbs, struct MBArtistInfo artist,
		  MBID* artist_mbidp)
{
  char * postfields;
  CURLcode ret;
  int ret_int;
  /*
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;
  char * header_info = NULL;
  int i;
  */

  postfields = build_artist_name_post_string(mbs->curl_handle, artist);

  add_type_to_artist(mbs->curl_handle, &postfields, artist);
  add_disambig_to_artist(mbs->curl_handle, &postfields, artist);
  add_urls_to_artist(mbs->curl_handle, &postfields, artist);
  add_relationships_to_artist(mbs->curl_handle, &postfields, artist);
  add_area_to_artist(mbs->curl_handle, &postfields, artist);

  ret_int = mb_send_post_query(mbs->curl_handle,
			       "https://musicbrainz.org/artist/create",
			       postfields, artist_mbidp);

  free((void *) postfields);
  return(ret_int);
}

int mb_update_artist(struct MBState * mbs, char musicbrainzID[],
		     struct MBArtistInfo artist)
{
  int ret_int;
  char * url ;
  char * resultpage = NULL;

  char * urlformat = "https://musicbrainz.org/artist/%s/edit";
  url = malloc(strlen(urlformat) + strlen(musicbrainzID));
  sprintf(url, urlformat, musicbrainzID);

  ret_int = mb_do_artist(mbs->curl_handle, url, artist);
  free(url);
}

int mb_add_place(struct MBState * mbs, struct MBPlace place,
		    MBID* place_mbidp)
{
  char * postfields;

  CURLcode ret;
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;

  char * subs;

  postfields = build_place_name_post_string(mbs->curl_handle, place);
  add_disambiguation_to_place(mbs->curl_handle, &postfields, place);
  /*
  add_dupe_to_place(mbs->curl_handle, &postfields, place);
  */
  add_type_to_place(mbs->curl_handle, &postfields, place);
  add_coordinates_to_place(mbs->curl_handle, &postfields, place);
  add_address_to_place(mbs->curl_handle, &postfields, place);
  add_area_id_to_place(mbs->curl_handle, &postfields, place);
  add_urls_to_place(mbs->curl_handle, &postfields, place);
  add_relationships_to_place(mbs->curl_handle, &postfields, place);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL,
			 "https://musicbrainz.org/place/create");

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(mbs->curl_handle);

  if(ret != CURLE_OK)
  {
    printf("in mb_add_place() - curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  if (redirect_url == NULL)
  {
    curl_easy_reset(mbs->curl_handle);
    free((void *) postfields);
    return(1);
  }

  for(subs = redirect_url; subs[0] != '\0'; subs = subs + 1) {
    if (strncmp(subs, "/place/", 7) == 0)
    {
      subs = subs + 7;
      break;
    }
  }

  strncpy(*place_mbidp, subs, MBID_length);
  (*place_mbidp)[MBID_length] = 0;
  curl_easy_reset(mbs->curl_handle);
  free((void *) postfields);
  return(0);
}

/*
currently assumes that event has one and only one venue
returns 0 on success (and undetected errors)
*/
int mb_do_event(CURL * ch, char * url, struct MBEventInfo event,
		 MBID* event_mbidp)
{
  char * postfields;

  /*
  MBID mbid_out;
  */

  CURLcode ret;
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;

  postfields = build_event_name_post_string(ch, event);
  add_disambig_to_event(ch, &postfields, event);
  add_setlist_to_event(ch, &postfields, event);

  add_type_to_event(ch, &postfields, event);
  add_begin_year_to_event(ch, &postfields, event);
  add_begin_month_to_event(ch, &postfields, event);
  add_begin_day_to_event(ch, &postfields, event);
  add_end_year_to_event(ch, &postfields, event);
  add_end_month_to_event(ch, &postfields, event);
  add_end_day_to_event(ch, &postfields, event);
  add_time_to_event(ch, &postfields, event);
  add_urls_to_event(ch, &postfields, event);
  add_relationships_to_event(ch, &postfields, event);

  ret = curl_easy_setopt(ch, CURLOPT_URL, url);

  ret = curl_easy_setopt(ch, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);

  ret = curl_easy_setopt(ch, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(ch);

  if(ret != CURLE_OK)
  {
    printf("in mb_do_event() - curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(ch, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(ch, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  /*
  printf("status_code is %ld\n", status_code);
  printf("redirect_url is %s\n", redirect_url);
  */

  if(status_code == 200 || status_code == 500)
  {
    free(resultpage);
    free((void *) postfields);
    return(1);
  }

  get_mbid_from_artist_url(redirect_url, event_mbidp);

  free(resultpage);
  free((void *) postfields);
  return(0);
}

/*
currently assumes that event has one and only one venue
*/
int mb_add_event(struct MBState * mbs, struct MBEventInfo event,
		 MBID* event_mbidp)
{
  char * postfields;

  int ret;

  postfields = build_event_name_post_string(mbs->curl_handle, event);
  add_disambig_to_event(mbs->curl_handle, &postfields, event);
  add_setlist_to_event(mbs->curl_handle, &postfields, event);

  add_type_to_event(mbs->curl_handle, &postfields, event);
  add_begin_year_to_event(mbs->curl_handle, &postfields, event);
  add_begin_month_to_event(mbs->curl_handle, &postfields, event);
  add_begin_day_to_event(mbs->curl_handle, &postfields, event);
  add_end_year_to_event(mbs->curl_handle, &postfields, event);
  add_end_month_to_event(mbs->curl_handle, &postfields, event);
  add_end_day_to_event(mbs->curl_handle, &postfields, event);
  add_time_to_event(mbs->curl_handle, &postfields, event);
  add_urls_to_event(mbs->curl_handle, &postfields, event);
  add_relationships_to_event(mbs->curl_handle, &postfields, event);

  ret = mb_send_post_query(mbs->curl_handle,
			   "https://musicbrainz.org/event/create",
			   postfields, event_mbidp);

  free((void *) postfields);
}

/*
currently assumes that event has one and only one venue
returns 0 on success (and undetected errors)
*/
int mb_update_event(struct MBState * mbs, char mbid[],
		    struct MBEventInfo event)
{
  char * url;

  int ret;
  MBID mbid_out;

  char * urlformat = "https://musicbrainz.org/event/%s/edit";
  url = malloc(strlen(urlformat) + strlen(mbid));
  sprintf(url, urlformat, mbid);

  ret = mb_do_event(mbs->curl_handle, url, event, &mbid_out);

  free(url);
  return(ret);

}

void mb_init_SeriesInfo(struct MBSeriesInfo * si)
{
  si->name = NULL;
  si->disambiguation = NULL;
  si->type = 0;
  si->ordering_type = 1;
}

char * build_series_name_post_string(CURL * ch, struct MBSeriesInfo series)
{
  return(build_form_chunk(ch, "edit-series.name", series.name));
}

int add_disambig_to_series(CURL * ch, char ** postfields,
			      struct MBSeriesInfo series)
{
  char* disambig = (series.disambiguation == NULL ? "": series.disambiguation);

  add_chunk_to_postfields4(ch, postfields, "edit-series.comment", disambig);

  return(0);
}

int add_type_to_series(CURL * ch, char ** postfields,
			      struct MBSeriesInfo series)
{
  if (series.type != 0)
    add_chunk_to_postfields5(ch, postfields,
			     "edit-series.type_id", series.type);

  return(0);
}

int add_ordering_type_to_series(CURL * ch, char ** postfields,
			      struct MBSeriesInfo series)
{
  add_chunk_to_postfields5(ch, postfields, "edit-series.ordering_type_id",
			   series.ordering_type);
  return(0);
  
}

int mb_do_series(CURL * ch, char * url, struct MBSeriesInfo series,
		 MBID * series_mbidp)
{
  char * postfields;

  int ret;

  postfields = build_series_name_post_string(ch, series);
  add_disambig_to_series(ch, &postfields, series);
  add_type_to_series(ch, &postfields, series);
  add_ordering_type_to_series(ch, &postfields, series);

  ret = mb_send_post_query(ch, url, postfields, series_mbidp);

  free((void *) postfields);
  return(0);
}

int mb_add_series(struct MBState * mbs, struct MBSeriesInfo series,
		  MBID * series_mbidp)
{
  int ret;

  ret = mb_do_series(mbs->curl_handle, "https://musicbrainz.org/series/create",
		     series, series_mbidp);

  return(ret);
}

void mb_init_Alias(struct MBAlias * alias)
{
  alias->name = NULL;
  alias->sort_name = NULL;
  alias->locale = NULL,
  alias->begin_year = 0;
  alias->begin_month = 0;
  alias->begin_day = 0;
  alias->end_year = 0;
  alias->end_month = 0;
  alias->end_day = 0;
  alias->type = 0;
}

char * build_alias_name_post_string(CURL * ch, struct MBAlias alias)
{
  return(build_form_chunk(ch, "edit-alias.name", alias.name));
}

int add_sort_name_to_alias(CURL * ch, char ** postfields,
			      struct MBAlias alias)
{
  char * sort_name = (alias.sort_name == NULL ? alias.name : alias.sort_name);

  add_chunk_to_postfields4(ch, postfields, "edit-alias.sort_name", sort_name);

  return(0);
}

int add_type_to_alias(CURL * ch, char ** postfields,
			      struct MBAlias alias)
{
  if (alias.type != 0)
    add_chunk_to_postfields5(ch, postfields,
			     "edit-alias.type_id", alias.type);

  return(0);
}

int mb_add_artist_alias(struct MBState * mbs, char musicbrainzID[],
			struct MBAlias alias)
{
  char * postfields;
  int ret;
  char * url ;

  char * urlformat = "https://musicbrainz.org/artist/%s/add-alias";
  url = malloc(strlen(urlformat) + strlen(musicbrainzID));
  sprintf(url, urlformat, musicbrainzID);

  postfields = build_alias_name_post_string(mbs->curl_handle, alias);
  add_sort_name_to_alias(mbs->curl_handle, &postfields, alias);
  add_type_to_alias(mbs->curl_handle, &postfields, alias);

  ret = mb_send_post_query(mbs->curl_handle, url, postfields, (MBID *) NULL);

  free((void *) postfields);

  free(url);
}
